#' simulate_data
#'
#' @param crop x/y dimension to crop to
#' @param tchange change point
#' @param tmax number of time points
#'
#' @return array of images xdim x ydim x time
#' @export
#'
#' @examples
simulate_data <- function(J=12,d=2,ti=80, stable=F, getTruth=F,Lsearch=NULL,noise="iid"){
  dx <- 2^(J)
  dy <- dx
  dat <- array(0,c(dx,dx,d+1,ti))
  if(!getTruth){
    # first calculate mean process - make it spatial to look nice
    gpm=gp(c(dx,dx),matern.specdens,c(1,0.1))
    simulate(gpm)
    mnval <- matrix(gpm$process,dx,dx)
  }
  
  # prepare some changes
  b1 <- 1
  
  # change patterns - small, medium, large
  p1 <- matrix(0,dx,dy)
  p1[floor(dx/5):floor(floor(dx/5)+dx/45),
     floor(dx/7):floor(floor(dx/7)+dx/38)] <- 1
  
  p2 <- matrix(0,dx,dy)
  p2[floor(3*dx/5):floor(floor(3*dx/5)+dx/24),
     floor(5*dx/7):floor(floor(5*dx/7)+dx/29)] <- 1
  
  p3 <- matrix(0,dx,dy)
  p3[floor(dx/8):floor(floor(dx/8)+dx/20),
     floor(7*dx/12):floor(floor(7*dx/12)+dx/17)] <- 1
  
  p4 <- matrix(0,dx,dy)
  p4[floor(8*dx/10):floor(floor(8*dx/10)+3*dx/20),
     floor(3*dx/12):floor(floor(3*dx/12)+dx/17)] <- 1
  
  p5 <- matrix(0,dx,dy)
  p5[floor(dx/2):floor(floor(dx/2)+dx/36),
     floor(dx/2):floor(floor(dx/2)+4*dx/31)] <- 1
  
  parr <- array(c(p1,p2,p3,p4,p5),c(dx,dx,5))
  
  tchange <- c(20,30,40,50,60)
  # now add some random noise
  ## iid
  
  if(getTruth){
    # change patterns - small, medium, large
    
      truth_mask_list <- lapply(1:ti,function(i){
      pmask <- apply(parr,c(1,2),function(x){sum(x[(tchange<=i)&(i<=(tchange+Lsearch))])})
      # pmask[pmask==0] <- NA
      # pmask_rast <- rast(ncols=ncol(pmask),nrows=nrow(pmask))
      # #ext(pmask_rast)[1:4] <- c(0,2^J,0,2^J)
      # values(pmask_rast) <- c(pmask)
      # pmask_buffer <- terra::buffer(pmask_rast,width=500000)
      # pmask_buffer
      # plot(pmask_buffer)
      return(pmask)
    })
    
    truth_mask <- array(unlist(truth_mask_list),c(dx,dy,ti))
    return(truth_mask)
  }
  
  if(!stable){
    epsilon <- rnorm(dx^2,0,0.5)
    for(di in 1:d){
      for(i in 1:ti){
        pmask <- apply(parr,c(1,2),function(x){sum(x[tchange<=i])})
        dat[,,di,i] <- mnval + b1*pmask + epsilon
        epsilon <- 0.4*epsilon+rnorm(dx^2,0,0.5)
      }
    }
  }else{
    epsilon <- rnorm(dx^2,0,0.5)
    for(di in 1:d){
      for(i in 1:ti){
        pmask <- apply(parr,c(1,2),function(x){sum(x)})
        dat[,,di,i] <- mnval + b1*pmask + epsilon
        epsilon <- 0.4*epsilon+rnorm(dx^2,0,0.5)
      }
    }
  }
  dat[,,d+1,] <- 0
  
  
  return(dat)
}





#' decomps_to_timeseries: takes a list of decompositions of images and 
#' restuctures
#'
#' @param dwt_list 
#'
#' @return
#' @export
#'
#' @examples
decomps_to_timeseries <- function(dwt_list){
  e.dwt <- dwt_list[[1]]
  nms <- names(dwt_list[[1]])
  dwt_struct <- lapply(nms,function(nm){
    array(unlist(lapply(dwt_list,function(x){
      x[[nm]]
    })),c(dim(e.dwt[[nm]]),length(dwt_list)))
  })
  names(dwt_struct) <- nms
  return(dwt_struct)
}


#' get_decomps
#'
#' @param arr an x by y by t array
#' @param J maximum decomp level
#'
#' @return array of wavelet coefficients for each time point (ns x t)
#' @export
#'
#' @examples
get_decomps <- function(arr,J=NULL){
  # choose level of decomp
  if(is.null(J)){
    J = log(min(dim(arr[,,1,1])),2)
  }
  # make empty dwt obj for getting dimensions
  e.dwt <- dwt.2d(array(0,dim(arr)[1:2]),wf="haar",J=J)
  index <- expand.grid(1:dim(arr)[3],1:dim(arr)[4])
  dwt_results <- lapply(1:nrow(index),function(x){
    x.dwt <- dwt.2d(arr[,,index[x,1],index[x,2]],wf="haar",J=J)
    return(x.dwt)
  })
  nms <- names(e.dwt)
  dwt_struct <- lapply(nms,function(nm){
    array(unlist(lapply(dwt_results,function(x){
      x[[nm]]
    })),c(dim(e.dwt[[nm]]),dim(arr)[3],dim(arr)[4]))
  })
  names(dwt_struct) <- nms
  
  return(dwt_struct)
}


#' run_roboBayes
#'
#' @param b.dwt.array 
#'
#' @return array of run lengths for each coefficient (tmax+1)x(tmax+1)xns
#' @export
#'
#' @examples
run_roboBayes <- function(b.dwt.list){
  nms <- names(b.dwt.list)
  tmax <- dim(b.dwt.list[[1]])[3]
  val <- lapply(b.dwt.list,function(b.dwt.array){
    s <- expand.grid(1:dim(b.dwt.array)[1],1:dim(b.dwt.array)[2])
    array(vapply(1:nrow(s),function(x){
      #print(x)
      priors <- list(B=matrix(0),V=matrix(0.1),Lambda=matrix(10),nu=3)
      mod <- roboBayes(b.dwt.array[s[x,1],s[x,2],],
                       rep(1,tmax),
                       par_inits = priors,
                       getR=T,
                       getOutliers=F,
                       getModels=F)
      return((mod$RFull))
    },FUN.VALUE = array(0,c(tmax+1,tmax+1))),c(tmax+1,tmax+1,dim(b.dwt.array)[1:2]))
    
  })
  names(val) <- nms
  return(val)
}

replace_na_with_last<-function(x,p=is.na,d=0)c(d,x)[cummax(seq_along(x)*(!p(x)))+1]

run_roboBayes_light <- function(signal,qa,priors,priors_adj,tvals,
                                Lsearch=30,truncRmin=35,qa_thresh=20){
  truncRmin <- max(truncRmin,Lsearch+8)
  doit <- function(){
    
  priorsi <- priors[[1]]
  datapts <- scale_data((signal),
                        0*c(qa)/100,
                        model_specs=priorsi,dates=tvals)
  Yvals <- datapts$y
  Xvals <- datapts$covariates
  good_ind <- which((qa<=qa_thresh) & rowMeans(!is.na(Yvals))==1 & rowMeans(!is.na(Xvals))==1)
  #Yvals[qa>=20,] <- NA
  #Xvals[qa>=20,] <- NA
  prs <- priorsi$priors[[1]]
  if(!is.null(priors_adj)){
    prs$Lambda <- priors_adj$Lambda
    prs$nu <- priors_adj$nu
    prs$V <- priors_adj$vscale*(prs$nu-d-1)/(priorsi$priors[[1]]$nu-d-1)*priorsi$priors[[1]]$V
  }
  
  if(length(good_ind)>=2){
    mod <- roboBayes(matrix(Yvals[good_ind,],length(good_ind),d),
                     Xvals[good_ind,],
                     lambda=1e8,
                     par_inits = prs,#priorsi$priors[[1]],
                     truncRmin=truncRmin,
                     Lsearch = Lsearch,
                     kt=1,
                     getR=F,
                     getOutliers=T,
                     Lm=7,
                     getModels=F,
                     outlier_mean=rep(0,ncol(Yvals)),
                     outlier_var = 2*(prs$nu-ncol(Yvals)-1)*diag(c(diag(prs$V)/prs$Lambda[1,1]),ncol(Yvals)))
    val <- rep(NA,length(datapts$y))
    val[good_ind] <- mod$lastLs
    val <- replace_na_with_last(val)
    #plot(val)
    #unique(c(mod$cpInds))
    return(val)
  }else{return(rep(0,nrow(datapts$y)))}
  }

  final_val <- try(
    doit(),
    silent = T
  )
  if (inherits(final_val, "try-error")) {
    warning("Something went wrong in roboBayes")
    return(rep(0,nrow(signal)))
  }else{
    return(final_val)
  }
  

}

wts_from_qa <- function(qa){
  val <- 1-qa
  val <- val*(1-0.1)+0.1
  return(val)
}

upsample <- function(A,b){
  # A is nxn, need it to be bnxbn
  Ax <- t(sapply(1:nrow(A),function(ax){
    t(rep(A[ax,],each=b))
  }))
  
  Axy <- sapply(1:ncol(Ax),function(ay){
    (rep(Ax[,ay],each=b))
    })
  return(Axy)
}

#' Title Convert spatial locations to vector value
#'
#' @param s1 
#' @param s2 
#' @param barb 
#'
#' @return
#' @export
#'
#' @examples
sub2ind <- function(si,m){
  c <- si[2]
  r <- si[1]
  ind = (c-1)*m + r
  return(ind)
}
ind2sub <- function(ind,m){
  r = ((ind-1) %% m) + 1
  c = floor((ind-1) / m) + 1
  return(c(r,c))
}



thresh_final <- function(im,threshold){
  return(im*(abs(im)>=threshold))
}

# 
wavelet_to_realspace_probchange <- function(component_probs,dg=1){
  # Pr(change) = 1-Pr(no change) = 1- prod((1-component_probs))
  if(dg==1){
    val <- 1-prod(1-component_probs)
  }
  if(dg==2){
    val <- 1-prod(1-component_probs)-sum(sapply(1:length(component_probs),function(x){component_probs[x]*prod(1-component_probs[-x])}))
  }
  if(dg==3){
    np <- length(component_probs)
    inds <- t(combn(length(component_probs),dg-1))
    val <- 1-prod(1-component_probs)-sum(sapply(1:length(component_probs),function(x){component_probs[x]*prod(1-component_probs[-x])}))-sum(
      sapply(1:nrow(inds),function(x){prod(component_probs[unlist(inds[x,])])*prod(1-component_probs[(1:np)[-inds[x,]]])}))
  }
  return(val)
}

count_wavelet_components <- function(component_probs,thresh){
  sum(component_probs > thresh)
}

testing <- function(component_probs,dg=2){
  np <- length(component_probs)
  cvals <- 1:np
  inds1 <- t(combn(length(component_probs),dg-1))
  inds1 <- as.matrix(inds1[nrow(inds1):1,])
  inds2 <- t(combn(length(component_probs),np-dg+1))
  val <- 1-prod(1-component_probs)-sum(rowProds(matrix(component_probs[inds1],nrow(inds1),ncol(inds1)))*rowProds(matrix(1-component_probs[inds2],nrow(inds2),ncol(inds2))))
  return(val)
  # matrix(cvals,nrow(inds2),nrow(inds2))[inds2]
  # val <- 1-prod(1-component_probs)-sum(sapply(1:length(component_probs),function(x){component_probs[x]*prod(1-component_probs[-x])}))-sum(
  #   sapply(1:nrow(inds),function(x){prod(component_probs[unlist(inds[x,])])*prod(1-component_probs[(1:np)[-inds[x,]]])}))
}

